'use strict';

const path = require('path');
const gulp = require('gulp');
const url = require('url');
const Handlebars = require('handlebars');
const jsdom = require('jsdom');
const glob = require('glob');
const read = require('read-file');
const reload = require('reload');
const jsonfile = require('jsonfile');
const fs = require('fs');
const crypto = require("crypto");
const copydir = require('copy-dir');
const request = require('request-json');

const regex = /{(.*)}/gm;
const subst = `:$1`;

module.exports.Helper = class Helper {
    constructor(config) {
        this.defaultLocale = config.defaultLocale || '';
        this.watchable = [];
        this.workingDir = config.workingDir || '.';
        this.components = config.components || {};
        this.routes = config.routes || {};
        this.layouts = config.layouts || {};
        this.resources = {
            translations: {},
        };
        this.componentsId = crypto.randomBytes(12).toString('hex');
        this.regex = /\$(\w+)/g; // api url variable replacement
    }

    getLayout(name, reload) {
        if (this.layouts.hasOwnProperty(name)) {
            const layout = this.layouts[name];
            if (reload) {
                layout.template = read.sync(layout.file, 'utf8');
            }
            return layout
        }
    }

    getComponent(name, reload) {
        if (this.components.hasOwnProperty(name)) {
            const component = this.components[name];
            if (reload) {
                this.createComponent(component.file)
            }
            return this.components[name]
        }
    }

    createComponent(filePath) {
        const component = {
            file: filePath,
            name: path.basename(filePath, '.html'),
            template: '',
        };

        const template = read.sync(component.file, 'utf8');
        const dom = new jsdom.JSDOM(template, {
            includeNodeLocations: true
        });
        const body = dom.window.document.head;
        for (let i = 0; i < body.children.length; i++) {
            let child = body.children[i];
            if (child.nodeName === "TEMPLATE") {
                const loc = dom.nodeLocation(child);
                const start = loc.startTag.endOffset;
                const end = loc.endOffset - start - 11;
                component.template = template.substr(start, end);
            } else if (child.nodeName === "SCRIPT") {
                component.script = child.innerHTML
            }
        }

        Handlebars.registerPartial(component.name, component.template);
        component.precompiled = Handlebars.precompile(component.template);
        this.components[component.name] = component;
    }

    createLayout(filePath) {
        const layout = {
            file: filePath,
            name: path.basename(filePath, '.html'),
        };
        layout.template = read.sync(layout.file, 'utf8');
        this.layouts[layout.name] = layout;
    }

    createResource(filePath) {
        const resource = read.sync(filePath, 'utf8');
        const jsonResource = JSON.parse(resource);
        if (jsonResource) {
            this.resources = Object.assign(this.resources || {}, jsonResource)
        }
    }

    parseLayout(pattern) {
        this.watchable.push({pattern: pattern, type: 'layout'});
        const files = glob.sync(pattern);
        for (let i = 0; i < files.length; i++) {
            this.createLayout(files[i])
        }
    }

    parse(pattern) {
        this.watchable.push({pattern: pattern, type: 'component'});
        const files = glob.sync(pattern);
        for (let i = 0; i < files.length; i++) {
            this.createComponent(files[i])
        }
    }

    parseResource(pattern) {
        this.watchable.push({pattern: pattern, type: 'resource'});
        const files = glob.sync(pattern);
        for (let i = 0; i < files.length; i++) {
            this.createResource(files[i])
        }
    }

    buildRoute(h, route, basePath, parents) {
        this.routeCount += 1;
        route.parents = parents;
        route.id = this.routeCount;
        let newPath = path.posix.join(basePath, route.path);

        // replace {param} with :param
        newPath = newPath.replace(regex, subst);

        if (route.path === '/') {
            newPath += '/'
        }
        if (!this.hasOwnProperty(newPath)) {
            h[newPath] = [];
        }
        if (parents) {
            h[newPath] = h[newPath].concat(parents)
        }
        if (!route.children || route.children.length === 0) {
            h[newPath].push(route)
        }
        if (route.children && route.children.length > 0) {
            let ps = parents || [];
            ps.push(route);
            for (let i = 0; i < route.children.length; i++) {
                h = this.buildRoute(h, route.children[i], newPath, ps)
            }
        }
        return h
    }

    changeDetected(watchable) {
        return change => {
            if (watchable.type === 'component') {
                this.createComponent(change.path)
            } else if (watchable.type === 'layout') {
                this.createLayout(change.path)
            } else if (watchable.type === 'resource') {
                this.createResource(change.path)
            }
            if (this.reload) {
                this.reload.reload()
            }
        }
    }

    setupExpress(app, opt) {
        // Serve components
        app.use('/' + this.componentsId + '.js', (req, res, next) => {
            const lang = req.query.lang || this.defaultLocale;
            res.send(this.packagedHandlebars(lang, true));
            //next()
        });

        this.routeCount = -1;
        this.handle = {}; // map of array of routes
        for (let i = 0; i < this.routes.length; i++) {
            this.handle = this.buildRoute(this.handle, this.routes[i], "/", [])
        }

        for (const p in this.handle) {
            if (this.handle.hasOwnProperty(p)) {
                const rs = this.handle[p];
                app.get(p, this.handleRoute(p, rs))
            }
        }

        if (opt) {
            if (opt.hasOwnProperty('liveReload') && opt.liveReload === true) {
                this.reload = reload(app);
                for (let i = 0; i < this.watchable.length; i++) {
                    const w = this.watchable[i];
                    gulp.watch(w.pattern, this.changeDetected(w))
                }
            }
        }

        // Language helper
        Handlebars.registerHelper("i18n", (k) => {
            return this.resources.translations[!!this.locale ? this.locale : this.defaultLocale][k]
        });

        // Definitions
        Handlebars.registerPartial('definitions', `<script src="/${this.componentsId}.js"></script>`);
    }

    handleRoute(p, rs) {
        return (req, res) => {
            return new Promise((resolve, reject) => {
                let context = {};
                if (req.params) {
                    context.query = req.params
                }
                if (req.query.hasOwnProperty('lang')) {
                    this.locale = req.query.lang || this.defaultLocale
                } else {
                    this.locale = this.defaultLocale
                }
                let layoutName = 'index';

                if (rs && rs.length > 0) {
                    const r = rs[0];
                    if (r && r.hasOwnProperty('layout') && r.layout.length > 0) {
                        layoutName = r.layout
                    }
                }

                const layout = this.getLayout(layoutName);
                const registeredPages = [];
                let resolvedApiUri = '';

                let outletName = 'router-outlet';
                for (let i = 0; i < rs.length; i++) {
                    const r = rs[i];
                    if (r.hasOwnProperty('redirect')) {
                        res.redirect(308, r.redirect);
                        return resolve()
                    }
                    if (r.hasOwnProperty('outlet') && !!r.outlet && r.outlet.length > 0) {
                        outletName = r.outlet
                    }
                    registeredPages.push(outletName);
                    Handlebars.registerPartial(outletName, this.components[r.component].template);
                    if (r.hasOwnProperty('api')) {
                        resolvedApiUri = r.api;
                        resolvedApiUri = resolvedApiUri.replace(this.regex, function (match, param) {
                            return match.replace('$' + param, req.params[param]);
                        });

                        let split = resolvedApiUri.split('?');
                        resolvedApiUri = split[0];
                        let resolvedQuery = '';
                        if (split.length > 1) {
                            resolvedQuery = split[1]
                        }

                        for (let k in req.query) {
                            if (req.query.hasOwnProperty(k)) {
                                if (resolvedQuery.length > 0) {
                                    resolvedQuery += '&'
                                }
                                resolvedQuery += k + '=' + req.query[k]
                            }
                        }

                        resolvedApiUri += '?' + resolvedQuery
                    }
                    if (r.hasOwnProperty('page')) {
                        context = Object.assign(context, r.page);
                    }
                }
                const fn = Handlebars.compile(layout.template);

                // add locale
                context['locale'] = this.locale;

                // get api data
                if (resolvedApiUri.length > 0) {
                    console.log('resolving ' + resolvedApiUri);
                    const client = request.createClient(resolvedApiUri);
                    client.get('', function (err, response, body) {
                        context.data = body;
                        context.json = JSON.stringify(context);
                        res.send(fn(context));
                        // remove pages from cache
                        for (let i = 0; i < registeredPages.length; i++) {
                            Handlebars.unregisterPartial(registeredPages[i])
                        }
                        resolve()
                    });
                } else {
                    context.json = JSON.stringify(context);
                    res.send(fn(context));
                    // remove pages from cache
                    for (let i = 0; i < registeredPages.length; i++) {
                        Handlebars.unregisterPartial(registeredPages[i])
                    }
                    resolve()
                }
            })
        }
    }

    packagedHandlebars(locale, isLocal) {
        const cArr = Object.keys(this.components).map(i => this.components[i]);
        if (isLocal) {
            const resources = Object.assign({}, this.resources || {});
            resources.translations = resources.translations ? resources.translations[locale] : {};
            return `${componentsLoader}${handlebarsScript}(function(){'use strict';const R=${JSON.stringify(!!resources ? resources : {})};window.translations=R.translations;window.storage=R.storage;Handlebars.registerHelper('i18n',(k)=>{return R.translations[k]});var template=Handlebars.template,templates=Handlebars.templates=Handlebars.templates || {};${cArr.map((component) => {
                return `Handlebars.partials['${component.name}'] = template(${component.precompiled});`
            }).join('')};WebComponents.waitFor=WebComponents.waitFor||function(p){return Promise.all([p].map(function(fn){return fn instanceof Function ? fn() : fn;}))};WebComponents.waitFor(()=>{return new Promise((res)=>{var cc=new function(){this.define=function(name,module){if(module&&module.hasOwnProperty('exports')){module.exports.prototype.template=function(c,o){return Handlebars.partials[name](c,o)};window.customElements.define(name,module.exports)}}};window['customComponents']=cc;${cArr.map((component) => {
                return component.script ? `cc.define('${component.name}',(function(){var module={};${component.script};return module})());` : ''
            }).join('')};res()})});var r=document.createElement('script');r.setAttribute('src','/reload/reload.js');document.head.appendChild(r)})();`
        } else {
            return [`${componentsLoader}${handlebarsScript}(function(){'use strict';var R=`, `;window.translations=R.translations;window.storage=R.storage;Handlebars.registerHelper('i18n',(k)=>{return R.translations[k]});var template=Handlebars.template,templates=Handlebars.templates=Handlebars.templates || {};${cArr.map((component) => {
                return `Handlebars.partials['${component.name}'] = template(${component.precompiled});`
            }).join('')};WebComponents.waitFor=WebComponents.waitFor||function(p){return Promise.all([p].map(function(fn){return fn instanceof Function ? fn() : fn;}))};WebComponents.waitFor(()=>{return new Promise((res)=>{var cc=new function(){this.define=function(name,module){if(module&&module.hasOwnProperty('exports')){module.exports.prototype.template=function(c,o){return Handlebars.partials[name](c,o)};window.customElements.define(name,module.exports)}}};window['customComponents']=cc;${cArr.map((component) => {
                return component.script ? `cc.define('${component.name}',(function(){var module={};${component.script};return module})());` : ''
            }).join('')};res()})})})();`]
        }
    }

    build(dist) {
        mkDirByPathSync(dist);
        mkDirByPathSync(dist + '/components');
        mkDirByPathSync(dist + '/public');

        // copy static files
        copydir.sync(this.workingDir + '/public', dist + '/public');
        // build manifest.json file

        const componentsVersion = crypto.randomBytes(6).toString('hex');
        const manifest = {
            'imports': [],
            'routes': this.routes,
            'defaultLocale': this.defaultLocale,
            'resources': this.resources,
            'componentsVersion': componentsVersion,
            'components': this.packagedHandlebars(),
        };
        for (let name in this.components) {
            if (this.components.hasOwnProperty(name)) {
                const newFilePath = 'components/' + name + '.component.html';

                // write file
                fs.writeFileSync(dist + '/' + newFilePath, this.components[name].template);

                manifest.imports.push({
                    'path': newFilePath,
                    'name': name,
                })
            }
        }

        // add definitions partial
        fs.writeFileSync(dist + "/definitions.html", `<script src="/${componentsVersion}.js?lang={{locale}}"></script>`);
        manifest.imports.push({
            'path': "definitions.html",
            'name': "definitions",
        });

        for (let name in this.layouts) {
            if (this.layouts.hasOwnProperty(name)) {
                const newFilePath = 'components/' + name + '.layout.html';

                // write file
                fs.writeFileSync(dist + '/' + newFilePath, this.layouts[name].template);

                manifest.imports.push({
                    'path': newFilePath,
                    'name': name,
                    'layout': true,
                })
            }
        }

        jsonfile.writeFileSync(dist + '/manifest.json', manifest)
    }
};

function mkDirByPathSync(targetDir, {isRelativeToScript = false} = {}) {
    const sep = path.sep;
    const initDir = path.isAbsolute(targetDir) ? sep : '';
    const baseDir = isRelativeToScript ? __dirname : '.';
    targetDir.split(sep).reduce((parentDir, childDir) => {
            const curDir = path.resolve(baseDir, parentDir, childDir);
            try {
                fs.mkdirSync(curDir);
            } catch (err) {
                if (err.code !== 'EEXIST') {
                    throw err;
                }
            }
            return curDir;
        },
        initDir
    )
}

const componentsLoader = `/**
 * @license
 * Copyright (c) 2018 The Polymer Project Authors. All rights reserved.
 * This code may only be used under the BSD style license found at http://polymer.github.io/LICENSE.txt
 * The complete set of authors may be found at http://polymer.github.io/AUTHORS.txt
 * The complete set of contributors may be found at http://polymer.github.io/CONTRIBUTORS.txt
 * Code distributed by Google as part of the polymer project is also
 * subject to an additional IP rights grant found at http://polymer.github.io/PATENTS.txt
 */

(function() {
  'use strict';

  /**
   * Basic flow of the loader process
   *
   * There are 4 flows the loader can take when booting up
   *
   * - Synchronous script, no polyfills needed
   *   - wait for \`DOMContentLoaded\`
   *   - run callbacks passed to \`waitFor\`
   *   - fire WCR event
   *
   * - Synchronous script, polyfills needed
   *   - document.write the polyfill bundle
   *   - wait on the \`load\` event of the bundle to batch Custom Element upgrades
   *   - wait for \`DOMContentLoaded\`
   *   - run callbacks passed to \`waitFor\`
   *   - fire WCR event
   *
   * - Asynchronous script, no polyfills needed
   *   - fire WCR event, as there could not be any callbacks passed to \`waitFor\`
   *
   * - Asynchronous script, polyfills needed
   *   - Append the polyfill bundle script
   *   - wait for \`load\` event of the bundle
   *   - batch Custom Element Upgrades
   *   - run callbacks pass to \`waitFor\`
   *   - fire WCR event
   */

  var polyfillsLoaded = false;
  var whenLoadedFns = [];
  var allowUpgrades = false;
  var flushFn;

  function fireEvent() {
    window.WebComponents.ready = true;
    document.dispatchEvent(new CustomEvent('WebComponentsReady', { bubbles: true }));
  }

  function batchCustomElements() {
    if (window.customElements && customElements.polyfillWrapFlushCallback) {
      customElements.polyfillWrapFlushCallback(function (flushCallback) {
        flushFn = flushCallback;
        if (allowUpgrades) {
          flushFn();
        }
      });
    }
  }

  function asyncReady() {
    batchCustomElements();
    ready();
  }

  function ready() {
    // bootstrap <template> elements before custom elements
    if (window.HTMLTemplateElement && HTMLTemplateElement.bootstrap) {
      HTMLTemplateElement.bootstrap(window.document);
    }
    polyfillsLoaded = true;
    runWhenLoadedFns().then(fireEvent);
  }

  function runWhenLoadedFns() {
    allowUpgrades = false;
    var done = function() {
      allowUpgrades = true;
      whenLoadedFns.length = 0;
      flushFn && flushFn();
    };
    return Promise.all(whenLoadedFns.map(function(fn) {
      return fn instanceof Function ? fn() : fn;
    })).then(function() {
      done();
    }).catch(function(err) {
      console.error(err);
    });
  }

  window.WebComponents = window.WebComponents || {
    ready: false,
    _batchCustomElements: batchCustomElements,
    waitFor: function(waitFn) {
      if (!waitFn) {
        return;
      }
      whenLoadedFns.push(waitFn);
      if (polyfillsLoaded) {
        runWhenLoadedFns();
      }
    }
  };

  // Feature detect which polyfill needs to be imported.
  var polyfills = [];
  if (!('attachShadow' in Element.prototype && 'getRootNode' in Element.prototype) ||
    (window.ShadyDOM && window.ShadyDOM.force)) {
    polyfills.push('sd');
  }
  if (!window.customElements || window.customElements.forcePolyfill) {
    polyfills.push('ce');
  }

  var needsTemplate = (function() {
    // no real <template> because no \`content\` property (IE and older browsers)
    var t = document.createElement('template');
    if (!('content' in t)) {
      return true;
    }
    // broken doc fragment (older Edge)
    if (!(t.content.cloneNode() instanceof DocumentFragment)) {
      return true;
    }
    // broken <template> cloning (Edge up to at least version 17)
    var t2 = document.createElement('template');
    t2.content.appendChild(document.createElement('div'));
    t.content.appendChild(t2);
    var clone = t.cloneNode(true);
    return (clone.content.childNodes.length === 0 ||
        clone.content.firstChild.content.childNodes.length === 0);
  })();

  // NOTE: any browser that does not have template or ES6 features
  // must load the full suite of polyfills.
  if (!window.Promise || !Array.from || !window.URL || !window.Symbol || needsTemplate) {
    polyfills = ['sd-ce-pf'];
  }

  if (polyfills.length) {
    var newScript = document.createElement('script');
    // Load it from the right place.
    var url = 'https://cdn.jsdelivr.net/npm/@webcomponents/webcomponentsjs@2.0.2/bundles/webcomponents-' + polyfills.join('-') + '.js';
    newScript.src = url;
    // if readyState is 'loading', this script is synchronous
    if (document.readyState === 'loading') {
      // make sure custom elements are batched whenever parser gets to the injected script
      newScript.setAttribute('onload', 'window.WebComponents._batchCustomElements()');
      document.write(newScript.outerHTML);
      document.addEventListener('DOMContentLoaded', ready);
    } else {
      newScript.addEventListener('load', function () {
        asyncReady();
      });
      newScript.addEventListener('error', function () {
        throw new Error('Could not load polyfill bundle' + url);
      });
      document.head.appendChild(newScript);
    }
  } else {
    polyfillsLoaded = true;
    if (document.readyState === 'complete') {
      fireEvent()
    } else {
      // this script may come between DCL and load, so listen for both, and cancel load listener if DCL fires
      window.addEventListener('load', ready);
      window.addEventListener('DOMContentLoaded', function() {
        window.removeEventListener('load', ready);
        ready();
      })
    }
  }
})();`;

const handlebarsScript = `/**!

 @license
 handlebars v4.0.11

Copyright (C) 2011-2017 by Yehuda Katz

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

*/
!function(a,b){"object"==typeof exports&&"object"==typeof module?module.exports=b():"function"==typeof define&&define.amd?define([],b):"object"==typeof exports?exports.Handlebars=b():a.Handlebars=b()}(this,function(){return function(a){function b(d){if(c[d])return c[d].exports;var e=c[d]={exports:{},id:d,loaded:!1};return a[d].call(e.exports,e,e.exports,b),e.loaded=!0,e.exports}var c={};return b.m=a,b.c=c,b.p="",b(0)}([function(a,b,c){"use strict";function d(){var a=new h.HandlebarsEnvironment;return n.extend(a,h),a.SafeString=j["default"],a.Exception=l["default"],a.Utils=n,a.escapeExpression=n.escapeExpression,a.VM=p,a.template=function(b){return p.template(b,a)},a}var e=c(1)["default"],f=c(2)["default"];b.__esModule=!0;var g=c(3),h=e(g),i=c(20),j=f(i),k=c(5),l=f(k),m=c(4),n=e(m),o=c(21),p=e(o),q=c(33),r=f(q),s=d();s.create=d,r["default"](s),s["default"]=s,b["default"]=s,a.exports=b["default"]},function(a,b){"use strict";b["default"]=function(a){if(a&&a.__esModule)return a;var b={};if(null!=a)for(var c in a)Object.prototype.hasOwnProperty.call(a,c)&&(b[c]=a[c]);return b["default"]=a,b},b.__esModule=!0},function(a,b){"use strict";b["default"]=function(a){return a&&a.__esModule?a:{"default":a}},b.__esModule=!0},function(a,b,c){"use strict";function d(a,b,c){this.helpers=a||{},this.partials=b||{},this.decorators=c||{},i.registerDefaultHelpers(this),j.registerDefaultDecorators(this)}var e=c(2)["default"];b.__esModule=!0,b.HandlebarsEnvironment=d;var f=c(4),g=c(5),h=e(g),i=c(9),j=c(17),k=c(19),l=e(k),m="4.0.11";b.VERSION=m;var n=7;b.COMPILER_REVISION=n;var o={1:"<= 1.0.rc.2",2:"== 1.0.0-rc.3",3:"== 1.0.0-rc.4",4:"== 1.x.x",5:"== 2.0.0-alpha.x",6:">= 2.0.0-beta.1",7:">= 4.0.0"};b.REVISION_CHANGES=o;var p="[object Object]";d.prototype={constructor:d,logger:l["default"],log:l["default"].log,registerHelper:function(a,b){if(f.toString.call(a)===p){if(b)throw new h["default"]("Arg not supported with multiple helpers");f.extend(this.helpers,a)}else this.helpers[a]=b},unregisterHelper:function(a){delete this.helpers[a]},registerPartial:function(a,b){if(f.toString.call(a)===p)f.extend(this.partials,a);else{if("undefined"==typeof b)throw new h["default"]('Attempting to register a partial called "'+a+'" as undefined');this.partials[a]=b}},unregisterPartial:function(a){delete this.partials[a]},registerDecorator:function(a,b){if(f.toString.call(a)===p){if(b)throw new h["default"]("Arg not supported with multiple decorators");f.extend(this.decorators,a)}else this.decorators[a]=b},unregisterDecorator:function(a){delete this.decorators[a]}};var q=l["default"].log;b.log=q,b.createFrame=f.createFrame,b.logger=l["default"]},function(a,b){"use strict";function c(a){return k[a]}function d(a){for(var b=1;b<arguments.length;b++)for(var c in arguments[b])Object.prototype.hasOwnProperty.call(arguments[b],c)&&(a[c]=arguments[b][c]);return a}function e(a,b){for(var c=0,d=a.length;c<d;c++)if(a[c]===b)return c;return-1}function f(a){if("string"!=typeof a){if(a&&a.toHTML)return a.toHTML();if(null==a)return"";if(!a)return a+"";a=""+a}return m.test(a)?a.replace(l,c):a}function g(a){return!a&&0!==a||!(!p(a)||0!==a.length)}function h(a){var b=d({},a);return b._parent=a,b}function i(a,b){return a.path=b,a}function j(a,b){return(a?a+".":"")+b}b.__esModule=!0,b.extend=d,b.indexOf=e,b.escapeExpression=f,b.isEmpty=g,b.createFrame=h,b.blockParams=i,b.appendContextPath=j;var k={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#x27;","\`":"&#x60;","=":"&#x3D;"},l=/[&<>"'\`=]/g,m=/[&<>"'\`=]/,n=Object.prototype.toString;b.toString=n;var o=function(a){return"function"==typeof a};o(/x/)&&(b.isFunction=o=function(a){return"function"==typeof a&&"[object Function]"===n.call(a)}),b.isFunction=o;var p=Array.isArray||function(a){return!(!a||"object"!=typeof a)&&"[object Array]"===n.call(a)};b.isArray=p},function(a,b,c){"use strict";function d(a,b){var c=b&&b.loc,g=void 0,h=void 0;c&&(g=c.start.line,h=c.start.column,a+=" - "+g+":"+h);for(var i=Error.prototype.constructor.call(this,a),j=0;j<f.length;j++)this[f[j]]=i[f[j]];Error.captureStackTrace&&Error.captureStackTrace(this,d);try{c&&(this.lineNumber=g,e?Object.defineProperty(this,"column",{value:h,enumerable:!0}):this.column=h)}catch(k){}}var e=c(6)["default"];b.__esModule=!0;var f=["description","fileName","lineNumber","message","name","number","stack"];d.prototype=new Error,b["default"]=d,a.exports=b["default"]},function(a,b,c){a.exports={"default":c(7),__esModule:!0}},function(a,b,c){var d=c(8);a.exports=function(a,b,c){return d.setDesc(a,b,c)}},function(a,b){var c=Object;a.exports={create:c.create,getProto:c.getPrototypeOf,isEnum:{}.propertyIsEnumerable,getDesc:c.getOwnPropertyDescriptor,setDesc:c.defineProperty,setDescs:c.defineProperties,getKeys:c.keys,getNames:c.getOwnPropertyNames,getSymbols:c.getOwnPropertySymbols,each:[].forEach}},function(a,b,c){"use strict";function d(a){g["default"](a),i["default"](a),k["default"](a),m["default"](a),o["default"](a),q["default"](a),s["default"](a)}var e=c(2)["default"];b.__esModule=!0,b.registerDefaultHelpers=d;var f=c(10),g=e(f),h=c(11),i=e(h),j=c(12),k=e(j),l=c(13),m=e(l),n=c(14),o=e(n),p=c(15),q=e(p),r=c(16),s=e(r)},function(a,b,c){"use strict";b.__esModule=!0;var d=c(4);b["default"]=function(a){a.registerHelper("blockHelperMissing",function(b,c){var e=c.inverse,f=c.fn;if(b===!0)return f(this);if(b===!1||null==b)return e(this);if(d.isArray(b))return b.length>0?(c.ids&&(c.ids=[c.name]),a.helpers.each(b,c)):e(this);if(c.data&&c.ids){var g=d.createFrame(c.data);g.contextPath=d.appendContextPath(c.data.contextPath,c.name),c={data:g}}return f(b,c)})},a.exports=b["default"]},function(a,b,c){"use strict";var d=c(2)["default"];b.__esModule=!0;var e=c(4),f=c(5),g=d(f);b["default"]=function(a){a.registerHelper("each",function(a,b){function c(b,c,f){j&&(j.key=b,j.index=c,j.first=0===c,j.last=!!f,k&&(j.contextPath=k+b)),i+=d(a[b],{data:j,blockParams:e.blockParams([a[b],b],[k+b,null])})}if(!b)throw new g["default"]("Must pass iterator to #each");var d=b.fn,f=b.inverse,h=0,i="",j=void 0,k=void 0;if(b.data&&b.ids&&(k=e.appendContextPath(b.data.contextPath,b.ids[0])+"."),e.isFunction(a)&&(a=a.call(this)),b.data&&(j=e.createFrame(b.data)),a&&"object"==typeof a)if(e.isArray(a))for(var l=a.length;h<l;h++)h in a&&c(h,h,h===a.length-1);else{var m=void 0;for(var n in a)a.hasOwnProperty(n)&&(void 0!==m&&c(m,h-1),m=n,h++);void 0!==m&&c(m,h-1,!0)}return 0===h&&(i=f(this)),i})},a.exports=b["default"]},function(a,b,c){"use strict";var d=c(2)["default"];b.__esModule=!0;var e=c(5),f=d(e);b["default"]=function(a){a.registerHelper("helperMissing",function(){if(1!==arguments.length)throw new f["default"]('Missing helper: "'+arguments[arguments.length-1].name+'"')})},a.exports=b["default"]},function(a,b,c){"use strict";b.__esModule=!0;var d=c(4);b["default"]=function(a){a.registerHelper("if",function(a,b){return d.isFunction(a)&&(a=a.call(this)),!b.hash.includeZero&&!a||d.isEmpty(a)?b.inverse(this):b.fn(this)}),a.registerHelper("unless",function(b,c){return a.helpers["if"].call(this,b,{fn:c.inverse,inverse:c.fn,hash:c.hash})})},a.exports=b["default"]},function(a,b){"use strict";b.__esModule=!0,b["default"]=function(a){a.registerHelper("log",function(){for(var b=[void 0],c=arguments[arguments.length-1],d=0;d<arguments.length-1;d++)b.push(arguments[d]);var e=1;null!=c.hash.level?e=c.hash.level:c.data&&null!=c.data.level&&(e=c.data.level),b[0]=e,a.log.apply(a,b)})},a.exports=b["default"]},function(a,b){"use strict";b.__esModule=!0,b["default"]=function(a){a.registerHelper("lookup",function(a,b){return a&&a[b]})},a.exports=b["default"]},function(a,b,c){"use strict";b.__esModule=!0;var d=c(4);b["default"]=function(a){a.registerHelper("with",function(a,b){d.isFunction(a)&&(a=a.call(this));var c=b.fn;if(d.isEmpty(a))return b.inverse(this);var e=b.data;return b.data&&b.ids&&(e=d.createFrame(b.data),e.contextPath=d.appendContextPath(b.data.contextPath,b.ids[0])),c(a,{data:e,blockParams:d.blockParams([a],[e&&e.contextPath])})})},a.exports=b["default"]},function(a,b,c){"use strict";function d(a){g["default"](a)}var e=c(2)["default"];b.__esModule=!0,b.registerDefaultDecorators=d;var f=c(18),g=e(f)},function(a,b,c){"use strict";b.__esModule=!0;var d=c(4);b["default"]=function(a){a.registerDecorator("inline",function(a,b,c,e){var f=a;return b.partials||(b.partials={},f=function(e,f){var g=c.partials;c.partials=d.extend({},g,b.partials);var h=a(e,f);return c.partials=g,h}),b.partials[e.args[0]]=e.fn,f})},a.exports=b["default"]},function(a,b,c){"use strict";b.__esModule=!0;var d=c(4),e={methodMap:["debug","info","warn","error"],level:"info",lookupLevel:function(a){if("string"==typeof a){var b=d.indexOf(e.methodMap,a.toLowerCase());a=b>=0?b:parseInt(a,10)}return a},log:function(a){if(a=e.lookupLevel(a),"undefined"!=typeof console&&e.lookupLevel(e.level)<=a){var b=e.methodMap[a];console[b]||(b="log");for(var c=arguments.length,d=Array(c>1?c-1:0),f=1;f<c;f++)d[f-1]=arguments[f];console[b].apply(console,d)}}};b["default"]=e,a.exports=b["default"]},function(a,b){"use strict";function c(a){this.string=a}b.__esModule=!0,c.prototype.toString=c.prototype.toHTML=function(){return""+this.string},b["default"]=c,a.exports=b["default"]},function(a,b,c){"use strict";function d(a){var b=a&&a[0]||1,c=s.COMPILER_REVISION;if(b!==c){if(b<c){var d=s.REVISION_CHANGES[c],e=s.REVISION_CHANGES[b];throw new r["default"]("Template was precompiled with an older version of Handlebars than the current runtime. Please update your precompiler to a newer version ("+d+") or downgrade your runtime to an older version ("+e+").")}throw new r["default"]("Template was precompiled with a newer version of Handlebars than the current runtime. Please update your runtime to a newer version ("+a[1]+").")}}function e(a,b){function c(c,d,e){e.hash&&(d=p.extend({},d,e.hash),e.ids&&(e.ids[0]=!0)),c=b.VM.resolvePartial.call(this,c,d,e);var f=b.VM.invokePartial.call(this,c,d,e);if(null==f&&b.compile&&(e.partials[e.name]=b.compile(c,a.compilerOptions,b),f=e.partials[e.name](d,e)),null!=f){if(e.indent){for(var g=f.split("\\n"),h=0,i=g.length;h<i&&(g[h]||h+1!==i);h++)g[h]=e.indent+g[h];f=g.join("\\n")}return f}throw new r["default"]("The partial "+e.name+" could not be compiled when running in runtime-only mode")}function d(b){function c(b){return""+a.main(e,b,e.helpers,e.partials,g,i,h)}var f=arguments.length<=1||void 0===arguments[1]?{}:arguments[1],g=f.data;d._setup(f),!f.partial&&a.useData&&(g=j(b,g));var h=void 0,i=a.useBlockParams?[]:void 0;return a.useDepths&&(h=f.depths?b!=f.depths[0]?[b].concat(f.depths):f.depths:[b]),(c=k(a.main,c,e,f.depths||[],g,i))(b,f)}if(!b)throw new r["default"]("No environment passed to template");if(!a||!a.main)throw new r["default"]("Unknown template object: "+typeof a);a.main.decorator=a.main_d,b.VM.checkRevision(a.compiler);var e={strict:function(a,b){if(!(b in a))throw new r["default"]('"'+b+'" not defined in '+a);return a[b]},lookup:function(a,b){for(var c=a.length,d=0;d<c;d++)if(a[d]&&null!=a[d][b])return a[d][b]},lambda:function(a,b){return"function"==typeof a?a.call(b):a},escapeExpression:p.escapeExpression,invokePartial:c,fn:function(b){var c=a[b];return c.decorator=a[b+"_d"],c},programs:[],program:function(a,b,c,d,e){var g=this.programs[a],h=this.fn(a);return b||e||d||c?g=f(this,a,h,b,c,d,e):g||(g=this.programs[a]=f(this,a,h)),g},data:function(a,b){for(;a&&b--;)a=a._parent;return a},merge:function(a,b){var c=a||b;return a&&b&&a!==b&&(c=p.extend({},b,a)),c},nullContext:l({}),noop:b.VM.noop,compilerInfo:a.compiler};return d.isTop=!0,d._setup=function(c){c.partial?(e.helpers=c.helpers,e.partials=c.partials,e.decorators=c.decorators):(e.helpers=e.merge(c.helpers,b.helpers),a.usePartial&&(e.partials=e.merge(c.partials,b.partials)),(a.usePartial||a.useDecorators)&&(e.decorators=e.merge(c.decorators,b.decorators)))},d._child=function(b,c,d,g){if(a.useBlockParams&&!d)throw new r["default"]("must pass block params");if(a.useDepths&&!g)throw new r["default"]("must pass parent depths");return f(e,b,a[b],c,0,d,g)},d}function f(a,b,c,d,e,f,g){function h(b){var e=arguments.length<=1||void 0===arguments[1]?{}:arguments[1],h=g;return!g||b==g[0]||b===a.nullContext&&null===g[0]||(h=[b].concat(g)),c(a,b,a.helpers,a.partials,e.data||d,f&&[e.blockParams].concat(f),h)}return h=k(c,h,a,g,d,f),h.program=b,h.depth=g?g.length:0,h.blockParams=e||0,h}function g(a,b,c){return a?a.call||c.name||(c.name=a,a=c.partials[a]):a="@partial-block"===c.name?c.data["partial-block"]:c.partials[c.name],a}function h(a,b,c){var d=c.data&&c.data["partial-block"];c.partial=!0,c.ids&&(c.data.contextPath=c.ids[0]||c.data.contextPath);var e=void 0;if(c.fn&&c.fn!==i&&!function(){c.data=s.createFrame(c.data);var a=c.fn;e=c.data["partial-block"]=function(b){var c=arguments.length<=1||void 0===arguments[1]?{}:arguments[1];return c.data=s.createFrame(c.data),c.data["partial-block"]=d,a(b,c)},a.partials&&(c.partials=p.extend({},c.partials,a.partials))}(),void 0===a&&e&&(a=e),void 0===a)throw new r["default"]("The partial "+c.name+" could not be found");if(a instanceof Function)return a(b,c)}function i(){return""}function j(a,b){return b&&"root"in b||(b=b?s.createFrame(b):{},b.root=a),b}function k(a,b,c,d,e,f){if(a.decorator){var g={};b=a.decorator(b,g,c,d&&d[0],e,f,d),p.extend(b,g)}return b}var l=c(22)["default"],m=c(1)["default"],n=c(2)["default"];b.__esModule=!0,b.checkRevision=d,b.template=e,b.wrapProgram=f,b.resolvePartial=g,b.invokePartial=h,b.noop=i;var o=c(4),p=m(o),q=c(5),r=n(q),s=c(3)},function(a,b,c){a.exports={"default":c(23),__esModule:!0}},function(a,b,c){c(24),a.exports=c(29).Object.seal},function(a,b,c){var d=c(25);c(26)("seal",function(a){return function(b){return a&&d(b)?a(b):b}})},function(a,b){a.exports=function(a){return"object"==typeof a?null!==a:"function"==typeof a}},function(a,b,c){var d=c(27),e=c(29),f=c(32);a.exports=function(a,b){var c=(e.Object||{})[a]||Object[a],g={};g[a]=b(c),d(d.S+d.F*f(function(){c(1)}),"Object",g)}},function(a,b,c){var d=c(28),e=c(29),f=c(30),g="prototype",h=function(a,b,c){var i,j,k,l=a&h.F,m=a&h.G,n=a&h.S,o=a&h.P,p=a&h.B,q=a&h.W,r=m?e:e[b]||(e[b]={}),s=m?d:n?d[b]:(d[b]||{})[g];m&&(c=b);for(i in c)j=!l&&s&&i in s,j&&i in r||(k=j?s[i]:c[i],r[i]=m&&"function"!=typeof s[i]?c[i]:p&&j?f(k,d):q&&s[i]==k?function(a){var b=function(b){return this instanceof a?new a(b):a(b)};return b[g]=a[g],b}(k):o&&"function"==typeof k?f(Function.call,k):k,o&&((r[g]||(r[g]={}))[i]=k))};h.F=1,h.G=2,h.S=4,h.P=8,h.B=16,h.W=32,a.exports=h},function(a,b){var c=a.exports="undefined"!=typeof window&&window.Math==Math?window:"undefined"!=typeof self&&self.Math==Math?self:Function("return this")();"number"==typeof __g&&(__g=c)},function(a,b){var c=a.exports={version:"1.2.6"};"number"==typeof __e&&(__e=c)},function(a,b,c){var d=c(31);a.exports=function(a,b,c){if(d(a),void 0===b)return a;switch(c){case 1:return function(c){return a.call(b,c)};case 2:return function(c,d){return a.call(b,c,d)};case 3:return function(c,d,e){return a.call(b,c,d,e)}}return function(){return a.apply(b,arguments)}}},function(a,b){a.exports=function(a){if("function"!=typeof a)throw TypeError(a+" is not a function!");return a}},function(a,b){a.exports=function(a){try{return!!a()}catch(b){return!0}}},function(a,b){(function(c){"use strict";b.__esModule=!0,b["default"]=function(a){var b="undefined"!=typeof c?c:window,d=b.Handlebars;a.noConflict=function(){return b.Handlebars===a&&(b.Handlebars=d),a}},a.exports=b["default"]}).call(b,function(){return this}())}])});`;